#!/usr/bin/env python
#
# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegweb main views."""

from django.http import Http404
from django.contrib import messages
from django.contrib.auth import login as auth_login
from django.contrib.auth import authenticate
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.shortcuts import render_to_response
from django.shortcuts import redirect
from django.template import RequestContext
from django.views.decorators.cache import never_cache
from django.views.decorators.http import require_POST
from django.contrib.auth.views import password_change as password_change_orig
from django.contrib.auth.views import password_change_done as password_change_done_orig

from pykeg.core import models
from pykeg.web.kegweb import forms
from pykeg.notification.forms import NotificationSettingsForm


@login_required
def account_main(request):
    context = RequestContext(request)
    context['user'] = request.user
    return render_to_response('account/index.html', context_instance=context)


@login_required
def edit_profile(request):
    context = RequestContext(request)
    user = request.user

    context['form'] = forms.ProfileForm(initial={'display_name': user.get_full_name()})

    if request.method == 'POST':
        form = forms.ProfileForm(request.POST, request.FILES)
        context['form'] = form
        if form.is_valid():
            if 'new_mugshot' in request.FILES:
                pic = models.Picture.objects.create(user=user)
                image = request.FILES['new_mugshot']
                pic.image.save(image.name, image)
                pic.save()
                user.mugshot = pic
            user.display_name = form.cleaned_data['display_name']
            user.save()
    return render_to_response('account/profile.html', context_instance=context)


@login_required
def invite(request):
    context = RequestContext(request)
    form = forms.InvitationForm()

    if not request.kbsite.can_invite(request.user):
        raise Http404('Cannot invite from this account.')

    if request.method == 'POST':
        form = forms.InvitationForm(request.POST)
        if form.is_valid():
            email = form.cleaned_data['email']
            invite = models.Invitation.objects.create(for_email=email,
                invited_by=request.user)
            invite.send()
            messages.success(request, 'Invitation mailed to ' + email)

    context['form'] = form
    return render_to_response('account/invite.html', context_instance=context)


@login_required
def notifications(request):
    # TODO(mikey): Dynamically add settings forms for other e-mail
    # backends (currently hardcoded to email backend).
    context = RequestContext(request)
    existing_settings = models.NotificationSettings.objects.get_or_create(user=request.user,
        backend='pykeg.notification.backends.email.EmailNotificationBackend')[0]
    if request.method == 'POST':
        form = NotificationSettingsForm(request.POST, instance=existing_settings)
        if form.is_valid():
            instance = form.save(commit=False)
            instance.user = request.user
            instance.backend = 'pykeg.notification.backends.email.EmailNotificationBackend'
            instance.save()
            messages.success(request, 'Settings updated')
            existing_settings = instance
    context['form'] = NotificationSettingsForm(instance=existing_settings)
    return render_to_response('account/notifications.html', context_instance=context)


@never_cache
def activate_account(request, activation_key):
    users = models.User.objects.filter(activation_key=activation_key)
    if users.count() != 1:
        raise Http404('No such activation key')
    user = users[0]

    assert not user.has_usable_password(), 'User already has a usable password'

    form = forms.ActivateAccountForm()
    if request.method == 'POST':
        form = forms.ActivateAccountForm(request.POST)
        if form.is_valid():
            cd = form.cleaned_data

            # Set the password and revoke the activation key.
            user.set_password(cd.get('password'))
            user.activation_key = None
            user.save()

            # Log the user in.
            user = authenticate(username=user.username, password=cd.get('password'))
            auth_login(request, user)
            if request.session.test_cookie_worked():
                request.session.delete_test_cookie()

            messages.success(request, 'Your account has been activated!')
            return redirect('kb-account-main')

    context = RequestContext(request)
    context['form'] = form
    return render_to_response('account/activate_account.html', context_instance=context)


@login_required
@require_POST
def regenerate_api_key(request):
    form = forms.RegenerateApiKeyForm(request.POST)
    if form.is_valid():
        key, is_new = models.ApiKey.objects.get_or_create(user=request.user)
        key.regenerate()
        key.save()
    return redirect('kb-account-main')


def password_change(request, *args, **kwargs):
    kwargs['template_name'] = 'account/password_change.html'
    kwargs['post_change_redirect'] = reverse('password_change_done')
    return password_change_orig(request, *args, **kwargs)


def password_change_done(request):
    return password_change_done_orig(request, 'account/password_change_done.html')


@login_required
def plugin_settings(request, plugin_name):
    plugin = request.plugins.get(plugin_name, None)
    if not plugin:
        raise Http404('Plugin "%s" not loaded' % plugin_name)

    view = plugin.get_user_settings_view()
    if not view:
        raise Http404('No user settings for this plugin')

    return view(request, plugin)



# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Webhook plugin for Kegbot."""

from pykeg.core.util import SuppressTaskErrors
from pykeg.plugin import plugin
from pykeg.proto import protolib

from . import forms
from . import tasks
from . import views

KEY_SITE_SETTINGS = 'settings'


class WebhookPlugin(plugin.Plugin):
    NAME = 'Web Hooks'
    SHORT_NAME = 'webhook'
    DESCRIPTION = 'Posts each system event to configured URLs'
    URL = 'http://kegbot.org'
    VERSION = '1.0.0'

    def get_admin_settings_view(self):
        return views.admin_settings

    def handle_new_events(self, events):
        for event in events:
            self.handle_event(event)

    def handle_event(self, event):
        self.logger.info('Handling new event: %s' % event.id)
        settings = self.get_site_settings()
        urls = settings.get('webhook_urls', '').strip().split()
        event_dict = protolib.ToDict(event, full=True)
        for url in urls:
            with SuppressTaskErrors(self.logger):
                tasks.webhook_post.delay(url, event_dict)

    ### Webhook-specific methods

    def get_site_settings_form(self):
        return self.datastore.load_form(forms.SiteSettingsForm, KEY_SITE_SETTINGS)

    def get_site_settings(self):
        return self.get_site_settings_form().initial

    def save_site_settings_form(self, form):
        self.datastore.save_form(form, KEY_SITE_SETTINGS)



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'User'
        db.create_table(u'core_user', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('password', self.gf('django.db.models.fields.CharField')(max_length=128)),
            ('last_login', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
            ('is_superuser', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('username', self.gf('django.db.models.fields.CharField')(unique=True, max_length=30)),
            ('display_name', self.gf('django.db.models.fields.CharField')(default='', max_length=127)),
            ('email', self.gf('django.db.models.fields.EmailField')(max_length=75, blank=True)),
            ('is_staff', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('is_active', self.gf('django.db.models.fields.BooleanField')(default=True)),
            ('date_joined', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
            ('mugshot', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='user_mugshot', null=True, on_delete=models.SET_NULL, to=orm['core.Picture'])),
            ('activation_key', self.gf('django.db.models.fields.CharField')(max_length=128, null=True, blank=True)),
        ))
        db.send_create_signal(u'core', ['User'])

        # Adding model 'Invitation'
        db.create_table(u'core_invitation', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('invite_code', self.gf('django.db.models.fields.CharField')(default='496f2e0d-36c7-4690-a243-f4f324b3c829', unique=True, max_length=255)),
            ('for_email', self.gf('django.db.models.fields.EmailField')(max_length=75)),
            ('invited_date', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('expires_date', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime(2014, 6, 5, 0, 0))),
            ('invited_by', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.User'], null=True, on_delete=models.SET_NULL)),
        ))
        db.send_create_signal(u'core', ['Invitation'])

        # Adding model 'KegbotSite'
        db.create_table(u'core_kegbotsite', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(default='default', unique=True, max_length=64)),
            ('server_version', self.gf('django.db.models.fields.CharField')(max_length=64, null=True)),
            ('is_setup', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('registration_id', self.gf('django.db.models.fields.TextField')(default='', max_length=128, blank=True)),
            ('volume_display_units', self.gf('django.db.models.fields.CharField')(default='imperial', max_length=64)),
            ('temperature_display_units', self.gf('django.db.models.fields.CharField')(default='f', max_length=64)),
            ('title', self.gf('django.db.models.fields.CharField')(default='My Kegbot', max_length=64)),
            ('background_image', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.Picture'], null=True, on_delete=models.SET_NULL, blank=True)),
            ('google_analytics_id', self.gf('django.db.models.fields.CharField')(max_length=64, null=True, blank=True)),
            ('session_timeout_minutes', self.gf('django.db.models.fields.PositiveIntegerField')(default=180)),
            ('privacy', self.gf('django.db.models.fields.CharField')(default='public', max_length=63)),
            ('registration_mode', self.gf('django.db.models.fields.CharField')(default='public', max_length=63)),
            ('timezone', self.gf('django.db.models.fields.CharField')(default='UTC', max_length=255)),
            ('check_for_updates', self.gf('django.db.models.fields.BooleanField')(default=True)),
        ))
        db.send_create_signal(u'core', ['KegbotSite'])

        # Adding model 'Device'
        db.create_table(u'core_device', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(default='Unknown Device', max_length=255)),
            ('created_time', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
        ))
        db.send_create_signal(u'core', ['Device'])

        # Adding model 'ApiKey'
        db.create_table(u'core_apikey', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.User'], null=True, blank=True)),
            ('device', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.Device'], null=True)),
            ('key', self.gf('django.db.models.fields.CharField')(default='3b59b5562ad9c0402f8e9565705a6949', unique=True, max_length=127)),
            ('active', self.gf('django.db.models.fields.BooleanField')(default=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('created_time', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
        ))
        db.send_create_signal(u'core', ['ApiKey'])

        # Adding model 'BeverageProducer'
        db.create_table(u'core_beverageproducer', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('country', self.gf('pykeg.core.fields.CountryField')(default='USA', max_length=3)),
            ('origin_state', self.gf('django.db.models.fields.CharField')(default='', max_length=128, null=True, blank=True)),
            ('origin_city', self.gf('django.db.models.fields.CharField')(default='', max_length=128, null=True, blank=True)),
            ('is_homebrew', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('url', self.gf('django.db.models.fields.URLField')(default='', max_length=200, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(default='', null=True, blank=True)),
            ('picture', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.Picture'], null=True, on_delete=models.SET_NULL, blank=True)),
            ('beverage_backend', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('beverage_backend_id', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
        ))
        db.send_create_signal(u'core', ['BeverageProducer'])

        # Adding model 'Beverage'
        db.create_table(u'core_beverage', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('producer', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.BeverageProducer'])),
            ('beverage_type', self.gf('django.db.models.fields.CharField')(default='beer', max_length=32)),
            ('style', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('picture', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.Picture'], null=True, blank=True)),
            ('vintage_year', self.gf('django.db.models.fields.DateField')(null=True, blank=True)),
            ('abv_percent', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
            ('calories_per_ml', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
            ('carbs_per_ml', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
            ('original_gravity', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
            ('specific_gravity', self.gf('django.db.models.fields.FloatField')(null=True, blank=True)),
            ('untappd_beer_id', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
            ('beverage_backend', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
            ('beverage_backend_id', self.gf('django.db.models.fields.CharField')(max_length=255, null=True, blank=True)),
        ))
        db.send_create_signal(u'core', ['Beverage'])

        # Adding model 'KegTap'
        db.create_table(u'core_kegtap', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=128)),
            ('description', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('current_keg', self.gf('django.db.models.fields.related.OneToOneField')(related_name='current_tap', null=True, on_delete=models.SET_NULL, to=orm['core.Keg'], blank=True, unique=True)),
            ('temperature_sensor', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.ThermoSensor'], null=True, on_delete=models.SET_NULL, blank=True)),
            ('sort_order', self.gf('django.db.models.fields.PositiveIntegerField')(default=0)),
        ))
        db.send_create_signal(u'core', ['KegTap'])

        # Adding model 'Controller'
        db.create_table(u'core_controller', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=128)),
            ('model_name', self.gf('django.db.models.fields.CharField')(max_length=128, null=True, blank=True)),
            ('serial_number', self.gf('django.db.models.fields.CharField')(max_length=128, null=True, blank=True)),
        ))
        db.send_create_signal(u'core', ['Controller'])

        # Adding model 'FlowMeter'
        db.create_table(u'core_flowmeter', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('controller', self.gf('django.db.models.fields.related.ForeignKey')(related_name='meters', to=orm['core.Controller'])),
            ('port_name', self.gf('django.db.models.fields.CharField')(max_length=128)),
            ('tap', self.gf('django.db.models.fields.related.OneToOneField')(blank=True, related_name='meter', unique=True, null=True, to=orm['core.KegTap'])),
            ('ticks_per_ml', self.gf('django.db.models.fields.FloatField')(default=5.4)),
        ))
        db.send_create_signal(u'core', ['FlowMeter'])

        # Adding unique constraint on 'FlowMeter', fields ['controller', 'port_name']
        db.create_unique(u'core_flowmeter', ['controller_id', 'port_name'])

        # Adding model 'FlowToggle'
        db.create_table(u'core_flowtoggle', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('controller', self.gf('django.db.models.fields.related.ForeignKey')(related_name='toggles', to=orm['core.Controller'])),
            ('port_name', self.gf('django.db.models.fields.CharField')(max_length=128)),
            ('tap', self.gf('django.db.models.fields.related.OneToOneField')(blank=True, related_name='toggle', unique=True, null=True, to=orm['core.KegTap'])),
        ))
        db.send_create_signal(u'core', ['FlowToggle'])

        # Adding unique constraint on 'FlowToggle', fields ['controller', 'port_name']
        db.create_unique(u'core_flowtoggle', ['controller_id', 'port_name'])

        # Adding model 'Keg'
        db.create_table(u'core_keg', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('type', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.Beverage'], on_delete=models.PROTECT)),
            ('keg_type', self.gf('django.db.models.fields.CharField')(default='half-barrel', max_length=32)),
            ('served_volume_ml', self.gf('django.db.models.fields.FloatField')(default=0)),
            ('full_volume_ml', self.gf('django.db.models.fields.FloatField')(default=0)),
            ('start_time', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
            ('end_time', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
            ('online', self.gf('django.db.models.fields.BooleanField')(default=True)),
            ('finished', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('description', self.gf('django.db.models.fields.CharField')(max_length=256, null=True, blank=True)),
            ('spilled_ml', self.gf('django.db.models.fields.FloatField')(default=0)),
            ('notes', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
        ))
        db.send_create_signal(u'core', ['Keg'])

        # Adding model 'Drink'
        db.create_table(u'core_drink', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('ticks', self.gf('django.db.models.fields.PositiveIntegerField')()),
            ('volume_ml', self.gf('django.db.models.fields.FloatField')()),
            ('time', self.gf('django.db.models.fields.DateTimeField')()),
            ('duration', self.gf('django.db.models.fields.PositiveIntegerField')(default=0, blank=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(related_name='drinks', to=orm['core.User'])),
            ('keg', self.gf('django.db.models.fields.related.ForeignKey')(related_name='drinks', on_delete=models.PROTECT, to=orm['core.Keg'])),
            ('session', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='drinks', null=True, on_delete=models.PROTECT, to=orm['core.DrinkingSession'])),
            ('shout', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('tick_time_series', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('picture', self.gf('django.db.models.fields.related.OneToOneField')(to=orm['core.Picture'], unique=True, null=True, on_delete=models.SET_NULL, blank=True)),
        ))
        db.send_create_signal(u'core', ['Drink'])

        # Adding model 'AuthenticationToken'
        db.create_table(u'core_authenticationtoken', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('auth_device', self.gf('django.db.models.fields.CharField')(max_length=64)),
            ('token_value', self.gf('django.db.models.fields.CharField')(max_length=128)),
            ('nice_name', self.gf('django.db.models.fields.CharField')(max_length=256, null=True, blank=True)),
            ('pin', self.gf('django.db.models.fields.CharField')(max_length=256, null=True, blank=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='tokens', null=True, to=orm['core.User'])),
            ('enabled', self.gf('django.db.models.fields.BooleanField')(default=True)),
            ('created_time', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
            ('expire_time', self.gf('django.db.models.fields.DateTimeField')(null=True, blank=True)),
        ))
        db.send_create_signal(u'core', ['AuthenticationToken'])

        # Adding unique constraint on 'AuthenticationToken', fields ['auth_device', 'token_value']
        db.create_unique(u'core_authenticationtoken', ['auth_device', 'token_value'])

        # Adding model 'DrinkingSession'
        db.create_table(u'core_drinkingsession', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('start_time', self.gf('django.db.models.fields.DateTimeField')()),
            ('end_time', self.gf('django.db.models.fields.DateTimeField')()),
            ('volume_ml', self.gf('django.db.models.fields.FloatField')(default=0)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=256, null=True, blank=True)),
        ))
        db.send_create_signal(u'core', ['DrinkingSession'])

        # Adding model 'ThermoSensor'
        db.create_table(u'core_thermosensor', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('raw_name', self.gf('django.db.models.fields.CharField')(max_length=256)),
            ('nice_name', self.gf('django.db.models.fields.CharField')(max_length=128)),
        ))
        db.send_create_signal(u'core', ['ThermoSensor'])

        # Adding model 'Thermolog'
        db.create_table(u'core_thermolog', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('sensor', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.ThermoSensor'])),
            ('temp', self.gf('django.db.models.fields.FloatField')()),
            ('time', self.gf('django.db.models.fields.DateTimeField')()),
        ))
        db.send_create_signal(u'core', ['Thermolog'])

        # Adding model 'Stats'
        db.create_table(u'core_stats', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('time', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
            ('stats', self.gf('pykeg.core.jsonfield.JSONField')()),
            ('drink', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.Drink'])),
            ('is_first', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(related_name='stats', null=True, to=orm['core.User'])),
            ('keg', self.gf('django.db.models.fields.related.ForeignKey')(related_name='stats', null=True, to=orm['core.Keg'])),
            ('session', self.gf('django.db.models.fields.related.ForeignKey')(related_name='stats', null=True, to=orm['core.DrinkingSession'])),
        ))
        db.send_create_signal(u'core', ['Stats'])

        # Adding unique constraint on 'Stats', fields ['drink', 'user', 'keg', 'session']
        db.create_unique(u'core_stats', ['drink_id', 'user_id', 'keg_id', 'session_id'])

        # Adding model 'SystemEvent'
        db.create_table(u'core_systemevent', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('kind', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('time', self.gf('django.db.models.fields.DateTimeField')()),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='events', null=True, to=orm['core.User'])),
            ('drink', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='events', null=True, to=orm['core.Drink'])),
            ('keg', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='events', null=True, to=orm['core.Keg'])),
            ('session', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='events', null=True, to=orm['core.DrinkingSession'])),
        ))
        db.send_create_signal(u'core', ['SystemEvent'])

        # Adding model 'Picture'
        db.create_table(u'core_picture', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('image', self.gf('django.db.models.fields.files.ImageField')(max_length=100)),
            ('time', self.gf('django.db.models.fields.DateTimeField')(default=datetime.datetime.now)),
            ('caption', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='pictures', null=True, to=orm['core.User'])),
            ('keg', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='pictures', null=True, on_delete=models.SET_NULL, to=orm['core.Keg'])),
            ('session', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='pictures', null=True, on_delete=models.SET_NULL, to=orm['core.DrinkingSession'])),
        ))
        db.send_create_signal(u'core', ['Picture'])

        # Adding model 'NotificationSettings'
        db.create_table(u'core_notificationsettings', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['core.User'])),
            ('backend', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('keg_tapped', self.gf('django.db.models.fields.BooleanField')(default=True)),
            ('session_started', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('keg_volume_low', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('keg_ended', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal(u'core', ['NotificationSettings'])

        # Adding unique constraint on 'NotificationSettings', fields ['user', 'backend']
        db.create_unique(u'core_notificationsettings', ['user_id', 'backend'])

        # Adding model 'PluginData'
        db.create_table(u'core_plugindata', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('plugin_name', self.gf('django.db.models.fields.CharField')(max_length=127)),
            ('key', self.gf('django.db.models.fields.CharField')(max_length=127)),
            ('value', self.gf('pykeg.core.jsonfield.JSONField')()),
        ))
        db.send_create_signal(u'core', ['PluginData'])

        # Adding unique constraint on 'PluginData', fields ['plugin_name', 'key']
        db.create_unique(u'core_plugindata', ['plugin_name', 'key'])


    def backwards(self, orm):
        # Removing unique constraint on 'PluginData', fields ['plugin_name', 'key']
        db.delete_unique(u'core_plugindata', ['plugin_name', 'key'])

        # Removing unique constraint on 'NotificationSettings', fields ['user', 'backend']
        db.delete_unique(u'core_notificationsettings', ['user_id', 'backend'])

        # Removing unique constraint on 'Stats', fields ['drink', 'user', 'keg', 'session']
        db.delete_unique(u'core_stats', ['drink_id', 'user_id', 'keg_id', 'session_id'])

        # Removing unique constraint on 'AuthenticationToken', fields ['auth_device', 'token_value']
        db.delete_unique(u'core_authenticationtoken', ['auth_device', 'token_value'])

        # Removing unique constraint on 'FlowToggle', fields ['controller', 'port_name']
        db.delete_unique(u'core_flowtoggle', ['controller_id', 'port_name'])

        # Removing unique constraint on 'FlowMeter', fields ['controller', 'port_name']
        db.delete_unique(u'core_flowmeter', ['controller_id', 'port_name'])

        # Deleting model 'User'
        db.delete_table(u'core_user')

        # Deleting model 'Invitation'
        db.delete_table(u'core_invitation')

        # Deleting model 'KegbotSite'
        db.delete_table(u'core_kegbotsite')

        # Deleting model 'Device'
        db.delete_table(u'core_device')

        # Deleting model 'ApiKey'
        db.delete_table(u'core_apikey')

        # Deleting model 'BeverageProducer'
        db.delete_table(u'core_beverageproducer')

        # Deleting model 'Beverage'
        db.delete_table(u'core_beverage')

        # Deleting model 'KegTap'
        db.delete_table(u'core_kegtap')

        # Deleting model 'Controller'
        db.delete_table(u'core_controller')

        # Deleting model 'FlowMeter'
        db.delete_table(u'core_flowmeter')

        # Deleting model 'FlowToggle'
        db.delete_table(u'core_flowtoggle')

        # Deleting model 'Keg'
        db.delete_table(u'core_keg')

        # Deleting model 'Drink'
        db.delete_table(u'core_drink')

        # Deleting model 'AuthenticationToken'
        db.delete_table(u'core_authenticationtoken')

        # Deleting model 'DrinkingSession'
        db.delete_table(u'core_drinkingsession')

        # Deleting model 'ThermoSensor'
        db.delete_table(u'core_thermosensor')

        # Deleting model 'Thermolog'
        db.delete_table(u'core_thermolog')

        # Deleting model 'Stats'
        db.delete_table(u'core_stats')

        # Deleting model 'SystemEvent'
        db.delete_table(u'core_systemevent')

        # Deleting model 'Picture'
        db.delete_table(u'core_picture')

        # Deleting model 'NotificationSettings'
        db.delete_table(u'core_notificationsettings')

        # Deleting model 'PluginData'
        db.delete_table(u'core_plugindata')


    models = {
        u'core.apikey': {
            'Meta': {'object_name': 'ApiKey'},
            'active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'created_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'device': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Device']", 'null': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'default': "'541f07b2788b4e33f152d05c440f183c'", 'unique': 'True', 'max_length': '127'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.User']", 'null': 'True', 'blank': 'True'})
        },
        u'core.authenticationtoken': {
            'Meta': {'unique_together': "(('auth_device', 'token_value'),)", 'object_name': 'AuthenticationToken'},
            'auth_device': ('django.db.models.fields.CharField', [], {'max_length': '64'}),
            'created_time': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'enabled': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'expire_time': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'nice_name': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'pin': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'token_value': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'tokens'", 'null': 'True', 'to': u"orm['core.User']"})
        },
        u'core.beverage': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Beverage'},
            'abv_percent': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'beverage_backend': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'beverage_backend_id': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'beverage_type': ('django.db.models.fields.CharField', [], {'default': "'beer'", 'max_length': '32'}),
            'calories_per_ml': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'carbs_per_ml': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'original_gravity': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'picture': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']", 'null': 'True', 'blank': 'True'}),
            'producer': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.BeverageProducer']"}),
            'specific_gravity': ('django.db.models.fields.FloatField', [], {'null': 'True', 'blank': 'True'}),
            'style': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'untappd_beer_id': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'vintage_year': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'})
        },
        u'core.beverageproducer': {
            'Meta': {'ordering': "('name',)", 'object_name': 'BeverageProducer'},
            'beverage_backend': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'beverage_backend_id': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'country': ('pykeg.core.fields.CountryField', [], {'default': "'USA'", 'max_length': '3'}),
            'description': ('django.db.models.fields.TextField', [], {'default': "''", 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_homebrew': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'origin_city': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'origin_state': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'picture': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'}),
            'url': ('django.db.models.fields.URLField', [], {'default': "''", 'max_length': '200', 'null': 'True', 'blank': 'True'})
        },
        u'core.controller': {
            'Meta': {'object_name': 'Controller'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model_name': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '128'}),
            'serial_number': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'})
        },
        u'core.device': {
            'Meta': {'object_name': 'Device'},
            'created_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'default': "'Unknown Device'", 'max_length': '255'})
        },
        u'core.drink': {
            'Meta': {'ordering': "('-time',)", 'object_name': 'Drink'},
            'duration': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'drinks'", 'on_delete': 'models.PROTECT', 'to': u"orm['core.Keg']"}),
            'picture': ('django.db.models.fields.related.OneToOneField', [], {'to': u"orm['core.Picture']", 'unique': 'True', 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'drinks'", 'null': 'True', 'on_delete': 'models.PROTECT', 'to': u"orm['core.DrinkingSession']"}),
            'shout': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'tick_time_series': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'ticks': ('django.db.models.fields.PositiveIntegerField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'drinks'", 'to': u"orm['core.User']"}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {})
        },
        u'core.drinkingsession': {
            'Meta': {'ordering': "('-start_time',)", 'object_name': 'DrinkingSession'},
            'end_time': ('django.db.models.fields.DateTimeField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {}),
            'volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'})
        },
        u'core.flowmeter': {
            'Meta': {'unique_together': "(('controller', 'port_name'),)", 'object_name': 'FlowMeter'},
            'controller': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'meters'", 'to': u"orm['core.Controller']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'port_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'tap': ('django.db.models.fields.related.OneToOneField', [], {'blank': 'True', 'related_name': "'meter'", 'unique': 'True', 'null': 'True', 'to': u"orm['core.KegTap']"}),
            'ticks_per_ml': ('django.db.models.fields.FloatField', [], {'default': '5.4'})
        },
        u'core.flowtoggle': {
            'Meta': {'unique_together': "(('controller', 'port_name'),)", 'object_name': 'FlowToggle'},
            'controller': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'toggles'", 'to': u"orm['core.Controller']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'port_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'tap': ('django.db.models.fields.related.OneToOneField', [], {'blank': 'True', 'related_name': "'toggle'", 'unique': 'True', 'null': 'True', 'to': u"orm['core.KegTap']"})
        },
        u'core.invitation': {
            'Meta': {'object_name': 'Invitation'},
            'expires_date': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime(2014, 6, 5, 0, 0)'}),
            'for_email': ('django.db.models.fields.EmailField', [], {'max_length': '75'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'invite_code': ('django.db.models.fields.CharField', [], {'default': "'82db659b-2873-4eab-96fb-256561188e79'", 'unique': 'True', 'max_length': '255'}),
            'invited_by': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.User']", 'null': 'True', 'on_delete': 'models.SET_NULL'}),
            'invited_date': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'})
        },
        u'core.keg': {
            'Meta': {'object_name': 'Keg'},
            'description': ('django.db.models.fields.CharField', [], {'max_length': '256', 'null': 'True', 'blank': 'True'}),
            'end_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'finished': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'full_volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg_type': ('django.db.models.fields.CharField', [], {'default': "'half-barrel'", 'max_length': '32'}),
            'notes': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'online': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'served_volume_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'spilled_ml': ('django.db.models.fields.FloatField', [], {'default': '0'}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Beverage']", 'on_delete': 'models.PROTECT'})
        },
        u'core.kegbotsite': {
            'Meta': {'object_name': 'KegbotSite'},
            'background_image': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Picture']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'}),
            'check_for_updates': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'google_analytics_id': ('django.db.models.fields.CharField', [], {'max_length': '64', 'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_setup': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'name': ('django.db.models.fields.CharField', [], {'default': "'default'", 'unique': 'True', 'max_length': '64'}),
            'privacy': ('django.db.models.fields.CharField', [], {'default': "'public'", 'max_length': '63'}),
            'registration_id': ('django.db.models.fields.TextField', [], {'default': "''", 'max_length': '128', 'blank': 'True'}),
            'registration_mode': ('django.db.models.fields.CharField', [], {'default': "'public'", 'max_length': '63'}),
            'server_version': ('django.db.models.fields.CharField', [], {'max_length': '64', 'null': 'True'}),
            'session_timeout_minutes': ('django.db.models.fields.PositiveIntegerField', [], {'default': '180'}),
            'temperature_display_units': ('django.db.models.fields.CharField', [], {'default': "'f'", 'max_length': '64'}),
            'timezone': ('django.db.models.fields.CharField', [], {'default': "'UTC'", 'max_length': '255'}),
            'title': ('django.db.models.fields.CharField', [], {'default': "'My Kegbot'", 'max_length': '64'}),
            'volume_display_units': ('django.db.models.fields.CharField', [], {'default': "'imperial'", 'max_length': '64'})
        },
        u'core.kegtap': {
            'Meta': {'ordering': "('sort_order', 'id')", 'object_name': 'KegTap'},
            'current_keg': ('django.db.models.fields.related.OneToOneField', [], {'related_name': "'current_tap'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Keg']", 'blank': 'True', 'unique': 'True'}),
            'description': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'sort_order': ('django.db.models.fields.PositiveIntegerField', [], {'default': '0'}),
            'temperature_sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.ThermoSensor']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'})
        },
        u'core.notificationsettings': {
            'Meta': {'unique_together': "(('user', 'backend'),)", 'object_name': 'NotificationSettings'},
            'backend': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg_ended': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keg_tapped': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'keg_volume_low': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'session_started': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.User']"})
        },
        u'core.picture': {
            'Meta': {'object_name': 'Picture'},
            'caption': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'image': ('django.db.models.fields.files.ImageField', [], {'max_length': '100'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Keg']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.DrinkingSession']"}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'pictures'", 'null': 'True', 'to': u"orm['core.User']"})
        },
        u'core.plugindata': {
            'Meta': {'unique_together': "(('plugin_name', 'key'),)", 'object_name': 'PluginData'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'key': ('django.db.models.fields.CharField', [], {'max_length': '127'}),
            'plugin_name': ('django.db.models.fields.CharField', [], {'max_length': '127'}),
            'value': ('pykeg.core.jsonfield.JSONField', [], {})
        },
        u'core.stats': {
            'Meta': {'unique_together': "(('drink', 'user', 'keg', 'session'),)", 'object_name': 'Stats'},
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_first': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'null': 'True', 'to': u"orm['core.Keg']"}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'null': 'True', 'to': u"orm['core.DrinkingSession']"}),
            'stats': ('pykeg.core.jsonfield.JSONField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'stats'", 'null': 'True', 'to': u"orm['core.User']"})
        },
        u'core.systemevent': {
            'Meta': {'ordering': "('-id',)", 'object_name': 'SystemEvent'},
            'drink': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.Drink']"}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'keg': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.Keg']"}),
            'kind': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'session': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.DrinkingSession']"}),
            'time': ('django.db.models.fields.DateTimeField', [], {}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'events'", 'null': 'True', 'to': u"orm['core.User']"})
        },
        u'core.thermolog': {
            'Meta': {'ordering': "('-time',)", 'object_name': 'Thermolog'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'sensor': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['core.ThermoSensor']"}),
            'temp': ('django.db.models.fields.FloatField', [], {}),
            'time': ('django.db.models.fields.DateTimeField', [], {})
        },
        u'core.thermosensor': {
            'Meta': {'object_name': 'ThermoSensor'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'nice_name': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'raw_name': ('django.db.models.fields.CharField', [], {'max_length': '256'})
        },
        u'core.user': {
            'Meta': {'object_name': 'User'},
            'activation_key': ('django.db.models.fields.CharField', [], {'max_length': '128', 'null': 'True', 'blank': 'True'}),
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'display_name': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '127'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'mugshot': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'user_mugshot'", 'null': 'True', 'on_delete': 'models.SET_NULL', 'to': u"orm['core.Picture']"}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        }
    }

    complete_apps = ['core']


# Copyright 2014 Bevbot LLC, All Rights Reserved
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Site backup and restore functions.

Backup file format is a zipfile with following contents:
    backup-name/         - top-level directory
       metadata.json     - backup metadata
       media/            - site media
       tables/           - per-table json records

The zipfile name takes the form <backup-name>-<sha1sum>.zip.

`backup-name` takes the form <site-title>-<YYMMDD-HHMMSS>, where
`site-title` is the slugified version of the current Kegbot site's
title.

Note that both the zipfile name and the backup file dirname are
informational and may change in a future version.

metadata.json consists of backup-related metadata:
  'created_time': ISO8601 timestamp at backup create
  'num_media_files': number of files in the 'media/' folder
  'num_tables': number of files in the 'tables/' folder
  'server_version': server version at backup create time
  'server_name': server name at backup create time
"""


import hashlib
import logging
import os
import sys
import isodate
import tempfile
import shutil
import json
import zipfile

from pykeg.core import models
from pykeg.core.util import get_version
from kegbot.util import kbjson

from django.core.files.storage import default_storage
from django.core import serializers
from django.db import transaction
from django.db import connection
from django.db.models import get_models
from django.utils import timezone
from django.utils.text import slugify

from south import models as south_models

logger = logging.getLogger(__name__)

# Tables to ignore when creating backups.  These tables will still be erased
# if they exist, which is important in the case of django_session.
BACKUP_SKIPPED_TABLE_NAMES = [
    u'django_admin_log',
    u'django_session',
    u'django_content_type',
    u'auth_group',
    u'auth_permission'
]

# These tables are not wiped on erase (and thus may be non-empty on restore).
ERASE_SKIPPED_TABLE_NAMES = [
    u'south_migrationhistory',
]

# App names to back up, needed for south migration.
APP_NAMES = [
    u'core',
]

# Whitelist of directories to include from storage backend.
MEDIA_WHITELIST = [
    u'pics'
]

# Metadata key names.
METADATA_FILENAME = 'metadata.json'
META_CREATED_TIME = 'created_time'
META_NUM_MEDIA_FILES = 'num_media_files'
META_NUM_TABLES = 'num_tables'
META_SERVER_VERSION = 'server_version'
META_SERVER_NAME = 'server_name'
META_BACKUP_FORMAT = 'backup_format'

BACKUPS_DIRNAME = 'backups'
TABLES_DIRNAME = 'tables'

# Current protocol.
BACKUP_FORMAT = 1


class BackupError(IOError):
    """Base backup exception."""


class AlreadyInstalledError(BackupError):
    """Attempt to restore against a non-pristine system."""


class InvalidBackup(BackupError):
    """Backup corrupt or not valid."""


class disable_foreign_key_checks(object):
    def __init__(self, connection):
        self.connection = connection

    def __enter__(self):
        if self.connection.vendor == 'mysql':
            cursor = self.connection.cursor()
            cursor.execute('SET foreign_key_checks = 0')

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.connection.vendor == 'mysql':
            cursor = self.connection.cursor()
            cursor.execute('SET foreign_key_checks = 1')
        return False


def read_metadata(zipfile):
    for info in zipfile.infolist():
        basename = os.path.basename(info.filename)
        if basename == METADATA_FILENAME:
            meta_file = zipfile.read(info.filename)
            return kbjson.loads(meta_file)
    return {}


def tbl(model):
    return str(model._meta.db_table)


def get_models_to_backup():
    """Returns models that should be saved during backup."""
    return [m for m in get_models() if tbl(m) not in BACKUP_SKIPPED_TABLE_NAMES]


def get_models_to_erase():
    """Returns models that should be erased [prior to restore]."""
    return [m for m in get_models_to_backup() if tbl(m) not in ERASE_SKIPPED_TABLE_NAMES]

get_models_to_restore = get_models_to_erase


def erase_model_table(model, must_truncate=False):
    table = tbl(model)
    cursor = connection.cursor()
    if connection.vendor == 'mysql':
        with disable_foreign_key_checks(connection):
            cursor.execute('TRUNCATE TABLE `{}`'.format(table))
            cursor.execute('ALTER TABLE `{}` AUTO_INCREMENT = 1'.format(table))
    elif connection.vendor == 'sqlite':
        cursor.execute('DELETE FROM {}'.format(table))
    else:
        if must_truncate:
            raise ValueError('Truncate on vendor "{}" unsupported'.format(connection.vendor))
        else:
            model.objects.all().delete()


@transaction.atomic
def create_backup_tree(date, storage, include_media=True):
    """Creates filesystem tree of backup data."""
    backup_dir = tempfile.mkdtemp()
    metadata = {}

    # Save databases.
    tables_dir = os.path.join(backup_dir, TABLES_DIRNAME)
    os.makedirs(tables_dir)

    all_models = get_models_to_backup()
    metadata[META_NUM_TABLES] = len(all_models)

    for model in all_models:
        table_name = model._meta.db_table
        output_filename = os.path.join(tables_dir, table_name + '.json')
        logger.debug('+++ Creating {}'.format(output_filename))
        with open(output_filename, 'w') as out:
            serializers.serialize('json', model.objects.all(), indent=2, stream=out)

    # Save stored media.
    metadata[META_NUM_MEDIA_FILES] = 0

    def add_files(storage, dirname, destdir):
        """Recursively copies all files in `dirname` to `destdir`."""
        subdirs, files = storage.listdir(dirname)
        for filename in files:
            full_filename = os.path.join(dirname, filename)
            output_filename = os.path.join(destdir, full_filename)
            output_dirname = os.path.dirname(output_filename)
            if not os.path.exists(output_dirname):
                os.makedirs(output_dirname)
            with storage.open(full_filename, 'r') as srcfile:
                with open(output_filename, 'w') as dstfile:
                    logger.debug('+++ Creating {}'.format(output_filename))
                    shutil.copyfileobj(srcfile, dstfile)
                    metadata[META_NUM_MEDIA_FILES] += 1
        for subdir in subdirs:
            add_files(storage, os.path.join((dirname, subdir)), destdir)

    if include_media:
        destdir = os.path.join(backup_dir, 'media')
        for media_dir in MEDIA_WHITELIST:
            if storage.exists(media_dir):
                add_files(storage, media_dir, destdir)
    else:
        logger.warning('Not including media.')

    # Store metadata file.
    metadata[META_SERVER_NAME] = models.KegbotSite.get().title
    metadata[META_SERVER_VERSION] = get_version()
    metadata[META_CREATED_TIME] = isodate.datetime_isoformat(date)
    metadata[META_BACKUP_FORMAT] = BACKUP_FORMAT
    metadata_filename = os.path.join(backup_dir, METADATA_FILENAME)
    with open(metadata_filename, 'w') as outfile:
        json.dump(metadata, outfile, sort_keys=True, indent=2)

    valid = False
    try:
        verify_backup_directory(backup_dir)
        valid = True
        return backup_dir
    finally:
        if not valid:
            shutil.rmtree(backup_dir)


def create_backup_zip(backup_dir, backup_name):
    """Creates a zipfile based on a tree created with `create_backup_tree()`."""
    zipfile_fileno, zipfile_path = tempfile.mkstemp()
    zipfile_fd = os.fdopen(zipfile_fileno, 'w')
    zf = zipfile.ZipFile(zipfile_fd, 'w')

    for dirname, subdirs, files in os.walk(backup_dir):
        for filename in files:
            full_path = os.path.join(dirname, filename)
            rel_path = os.path.relpath(full_path, backup_dir)
            archive_path = os.path.join(backup_name, rel_path)
            zf.write(full_path, archive_path)
            logger.debug('Added to zip: {}'.format(archive_path))

    zf.close()
    return zipfile_path


def backup(storage=default_storage, include_media=True):
    """Creates a new backup archive.

    Returns:
        A path to the stored zip file, in terms of the current storage class.
    """
    date = timezone.now()
    site_slug = slugify(models.KegbotSite.get().title)
    date_str = date.strftime('%Y%m%d-%H%M%S')
    backup_name = '{slug}-{date}'.format(slug=site_slug, date=date_str)

    backup_dir = create_backup_tree(date=date, storage=storage, include_media=include_media)
    try:
        temp_zip = create_backup_zip(backup_dir, backup_name)
        try:
            # Append sha1sum.
            sha1 = hashlib.sha1()
            with open(temp_zip, 'rb') as f:
                for chunk in iter(lambda: f.read(2 ** 20), b''):
                    sha1.update(chunk)
            digest = sha1.hexdigest()
            saved_zip_name = os.path.join(BACKUPS_DIRNAME,
                '{}-{}.zip'.format(backup_name, digest))
            with open(temp_zip, 'r') as temp_zip_file:
                ret = storage.save(saved_zip_name, temp_zip_file)
                return ret
        finally:
            os.unlink(temp_zip)
    finally:
        shutil.rmtree(backup_dir)


def extract_backup(backup_file):
    assert sys.version_info[:3] >= (2, 7, 4), "Unsafe to extract ZIPs in this Python version"
    backup_dir = tempfile.mkdtemp()
    logger.debug('Extracting backup to {}'.format(backup_dir))
    zf = zipfile.ZipFile(backup_file, mode='r')
    zf.extractall(path=backup_dir)
    return backup_dir


def verify_backup_directory(backup_dir):
    metadata_file = os.path.join(backup_dir, METADATA_FILENAME)
    if not os.path.exists(metadata_file):
        raise InvalidBackup('Metadata file does not exist')

    metadata = kbjson.loads(open(metadata_file, 'r').read())
    format = metadata.get(META_BACKUP_FORMAT)
    if format != BACKUP_FORMAT:
        raise InvalidBackup('Unsupported backup format: {}'.format(format))

    for model in get_models_to_restore():
        table_json = os.path.join(backup_dir, TABLES_DIRNAME, tbl(model) + '.json')
        if not os.path.exists(table_json):
            raise InvalidBackup('Backup missing file {}'.format(table_json))


def restore_tables(backup_dir):
    """Loads database tables from `backup_dir`."""
    for model in get_models_to_restore():
        if model.objects.all().count() > 0:
            raise AlreadyInstalledError('Table "{}" is not empty, cannot restore. '
                'Run "kegbot erase" first.'.format(model._meta.db_table))
    tables_dir = os.path.join(backup_dir, TABLES_DIRNAME)

    with disable_foreign_key_checks(connection):
        for model in get_models_to_restore():
            table = tbl(model)
            table_file = os.path.join(tables_dir, table + '.json')
            logger.debug('Restoring table {}'.format(table))
            with open(table_file, 'r') as table_fp:
                data = table_fp.read()
                objects = serializers.deserialize('json', data)
                for obj in objects:
                    obj.save()


def restore_media(backup_dir, storage):
    media_dir = os.path.join(backup_dir, 'media')

    for dirname, subdirs, files in os.walk(media_dir):
        for filename in files:
            full_path = os.path.join(dirname, filename)
            rel_path = os.path.relpath(full_path, media_dir)
            with open(full_path, 'r') as data:
                logger.debug('+++ Restoring file {}'.format(rel_path))
                storage.save(rel_path, data)


def check_app_migrations(app_name, backup_migrations):
    """Asserts that latest migration for `app_name` match those saved in
       `backup_migrations`.
    """
    latest_backup = latest_db = None
    for m in backup_migrations:
        if m.object.app_name == app_name:
            if not latest_backup or latest_backup.migration < m.object.migration:
                latest_backup = m.object

    qs = south_models.MigrationHistory.objects.filter(app_name=app_name)
    qs = qs.order_by('-migration')
    if qs:
        latest_db = qs[0]

    if latest_backup is None:
        raise InvalidBackup('No migrations in backup for app "{}"'.format(app_name))
    if latest_db is None:
        raise InvalidBackup('No migrations in database for app "{}"'.format(app_name))

    logger.debug('App "{}" migrations: backup={} db={}'.format(
        app_name, latest_backup.migration, latest_db.migration))

    if latest_backup.migration != latest_db.migration:
        raise InvalidBackup('Backup migration mismatch (db={}).  Try "kegbot migrate {} {}" first.'.format(
            latest_db.migration, app_name, latest_backup.migration))


def check_migrations(backup_dir):
    tables_dir = os.path.join(backup_dir, TABLES_DIRNAME)
    migration_filename = os.path.join(tables_dir, 'south_migrationhistory.json')
    with open(migration_filename, 'r') as migration_fp:
        backup_migrations = list(serializers.deserialize('json', migration_fp.read()))

    for app_name in APP_NAMES:
        logger.debug('Checking migrations for app {}'.format(app_name))
        check_app_migrations(app_name, backup_migrations)


def restore_from_directory(backup_dir, storage=default_storage):
    logger.info('Restoring from {} ...'.format(backup_dir))
    verify_backup_directory(backup_dir)
    with transaction.atomic():
        check_migrations(backup_dir)
        restore_tables(backup_dir)
        restore_media(backup_dir, storage)
    logger.info('Restore completed successfully.')


def restore(backup_file):
    """Restores from a backup zipfile.

    The site must be erased prior to running restore.
    """
    logger.info('Restoring from {} ...'.format(backup_file))
    unzipped_dir = extract_backup(backup_file)
    try:
        dirs = os.listdir(unzipped_dir)
        if len(dirs) != 1:
            raise InvalidBackup('Expected exactly one directory in archive.')
        backup_dir = os.path.join(unzipped_dir, dirs[0])
        restore_from_directory(backup_dir)
    finally:
        shutil.rmtree(unzipped_dir)


def erase(storage=default_storage):
    """Erases ALL site data (database tables, media files)."""
    logger.info('Erasing tables ...')
    with transaction.atomic():
        with disable_foreign_key_checks(connection):
            for model in get_models_to_erase():
                logger.debug('--- Erasing {}'.format(tbl(model)))
                erase_model_table(model, must_truncate=True)

    def delete_files(dirname):
        subdirs, files = storage.listdir(dirname)
        for filename in files:
            full_name = os.path.join(dirname, filename)
            logger.debug('Deleting file: {}'.format(full_name))
            storage.delete(full_name)
        for subdir in subdirs:
            delete_files(dirname)

    logger.info('Erasing media ...')
    for media_dir in MEDIA_WHITELIST:
        if storage.exists(media_dir):
            delete_files(media_dir)

    logger.info('Done.')
